// --- [ ir_sla_receive.c ] 18T -------- ----------------------------------------
//       tab = 3, 24.01.09
// 
// fr den IR-Slave genderte Routine zum Empfangen von 1 Wort via Manchester-Codierung auf 36 KHz
// nicht kompatibel mit RC5-Standard !!

// Benutzt den Timer0 zur Bestimmung der Impulsdauer und fr eine Zeitbegrenzung (Timer0_ovf)

// Arbeitet interruptgesteuert:
// Wartet auf die fallende Flanke eines Startbits, dann wird der Timer gestartet mit TCNT0 = 0
// Das Startbit ist 1/2 Takt lang.
// Bei jedem folgenden Interrupt wird die zeitliche Distanz zum
// Vorgnger getestet:
// bei 1/2 Taktabstand liegt nur Flankenwechsel vor,
// bei 1/1 Taktabstand liegt der relevante Flankenwechsel in Taktmitte vor,
// bei fallender Flanke wurde eine '1', bei steigender Flanke eine '0' bertragen.

// Die eingehenden Bits werden von rechts nach links in die Variable ir_in geschoben.

// Nach 16 Bits ist das Ende erreicht.
// Wenn weniger Bits eintreffen, lst der Timer0_OVF nach 1.5 Takten ohne INT0 einen Fehler aus,
// treffen mehr Bits ein, dann wird das Ergebnis berschrieben und lst dann
// wenn nicht zufllig wieder genau 16 Bit eintreffen - einen Fehler aus.

// Die Dauer eines Bits betrgt 18 Takte bei 36KHz low sowie
// 18 Takte high: (18 + 18) / 36000 = 36 / 36000 =  1000us
// bei Prescale des Timers von '64' 64/8.000.000 =  8us/Tick
// der TCNTO erhht sich innerhalb eines ganzen Taktes um 1000us/8us = 125 

// Achtung Verstndishrde:
// wenn das zuletzt eingehende Bit eine '1' ist, dann folgt auf die fallende
// Flanke noch eine steigende (das Signal geht in den Ruhezustand HIGH) zurck.
// Diese Flanke lst einen Interrupt aus, der als Start fr ein folgendes Byte
// zhlt ... 
// Um dies Problem zu lsen, wird nach dem letzten signifikanten Levelwechsel
// der INT0 Sense Control von 'any logic change' auf 'falling edge' geschaltet,
// dadurch wird ein folgende steigende Flanke ignoriert.

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include "ir_slave.h"

// lokale variable
volatile uint16_t ir_in;					// 16Bit Daten
volatile uint8_t  ir_start;				// Flag fr den Status des Empfngers
volatile uint8_t  ir_error;				// Fehlercode
volatile uint8_t  ir_ok;					// zeigt an, dass gltige Daten vorliegen
volatile uint8_t  bitcount;				// Zhler fr eingegangene Bits

// ----------------------------------------------------------------------------
// Hier wird der IR-Empfnger konfiguriert.
// Der Timer0 luft mit Prescaler 64 -> 64 / 8.000.000  -> 8us/Tick
void IR_init(void)
{
TIMSK = (1<<TOIE0);							// Timer0 Overflow Interrupt Enable
TCCR0A = 0;										// Timer0 als reiner Counter
PCMSK = (1<<PCINT4);							// Pin Change an PinB.4 aktivieren
GIMSK = (1<<PCIE);							// Pin Change Interrupt enablen
sei();											// INterrupts aktivieren
}

// ----------------------------------------------------------------------------
// exportiert, ob Daten empfangen wurden, liefert TRUE, wenn 16 Bit eingegangen sind
uint8_t IR_complete(void)
{
return ir_ok;
}

// ----------------------------------------------------------------------------
// exportiert, ob Daten empfangen werden, liefert TRUE, wenn das Startbit eingegangen ist
uint8_t IR_busy(void)
{
return ir_start;
}

// ----------------------------------------------------------------------------
// exportiert das empfangene Wort und lscht Flag ir_ok, 
// damit neue Daten erkannt werden knnen
uint16_t IR_read(void)
{
ir_ok = 0;
return ir_in; 
}

// ----------------------------------------------------------------------------
// beendet eine bertragung, wenn 1.5 Takte lang kein PCINT0 aufgetreten ist
ISR(TIM0_OVF_vect)
{
TCCR0B = T0_STOP;						// Timer0 stoppen
ir_start = 0;							// Neustart erzwingen
ir_error = 4;							// Fehlercode 
}

// ----------------------------------------------------------------------------
// wird bei Flankenwechsel an Pin.B4 aufgerufen
ISR(PCINT0_vect)
{
static volatile uint8_t only_falling_edge = TRUE;
register uint8_t distance;
distance = TCNT0;
													// Der 1. Impuls kommt vom Wakeup (falls gesendet)
													// diese fhrt aber mit 5ms Wartezeit zu Timer0_OVF und Ende
													// und damit wieder an diese Stelle
													// Der 2. 1.Impuls kommt dann vom Startbit, das 1.5 Takte lang ist.
											
if (ir_start == 0)							// Fallende Flanke des Weckbits bzw. des Startbits
	{												// der Weckruf ndert nur ir_start und INitialisiert die Variablen
	ir_start = 1;								// Flag
	bitcount = 0;								// Zhler fr die eingegangenen Bits	
	ir_in = 0;									// Variable zum Einsammeln der Bits
	ir_ok = 0;									// wenn die 16 Bit zusammen sind, dann wird ir_ok gesetzt
	ir_error = 0;								// zeigt an, an welcher Stelle sich ein Fehler ereignet hat (timing)
	only_falling_edge = FALSE;				// falling und rising edge werden bearbeitet	
	
	TCNT0 = 0;									// Counter zurcksetzen um die Lnge des Startbits zu beschrnken
													// fr Overflow nach 256 * 16us = 4096us
	TCCR0B = T0_START;						// Timer starten, ein Overflow setzt ir_start wieder auf '0'
	}

else if (ir_start == 1)						// die nachfolgende steigende Flanke des Startbits
	{												// das Startbit 1/2 Takt lang 
	if ((distance < (hT - TO)) || (distance > (hT + TO)))		
		{											// das war dann kein richtiges Startbit ...
		ir_start = 0;							// neuer Start von Vorne beginnen
		TCCR0B = T0_STOP;						// Timer abschalten
		only_falling_edge = TRUE;			// eine rising edge wird nicht weiter bearbeitet
		}

	else											// das Startbit war korrekt
		{											// die steigende Flanke des Startbits startet den Timer
		ir_start = 2;							// damit diese Initialisierung nur einmal am Start ausgefhrt wird
		TCNT0 = T0_PRELOAD;					// Counter zurcksetzen fr Overflow nach ca. 1.5 Bitzeit
		}
	}
	
else												// if ir_start > 1
	{

// als Prescale ist 64 eingestellt, TCCTN0 zahlt je 8.000.000 / 64 = 8us um 1 hoch
// die Bitzeit = 36/36000 = 1000us -> 1000/8 = 125, halbe Bitzeit 62
													// wenn 1/2 Takt Abstand, dann liegt nur ein Flankenwechsel
													// zwischen zwei gleichen Bytes vor
											
	if (distance < (hT - TO + T0_PRELOAD))// zu kurz fr halben Takt
		{
		ir_start = 0;							// Neustart erzwingen
		ir_error = 1;							// Fehlerursache dokumentieren
		}
													// es handelt sich um den Flankenwechsel zwischen zwei Bits
	else if (distance < (hT + TO + T0_PRELOAD))
		{
		return;									// hier gibt es nichts zu tun 
		}

	else if (distance < (gT - TO + T0_PRELOAD))// zu kurz fr ganzen Takt
		{
		ir_start = 0;							// Neustart erzwingen
		ir_error = 2;							// Fehlerursache dokumentieren
		}
													// zu lang fr einen ganzen Takt
	else if (distance > (gT + TO + T0_PRELOAD)) 
		{
		ir_start = 0;							// Neustart erzwingen
		ir_error = 3;							// Fehlerursache dokumentieren
		}

	else											// gltiger Taktabstand fr ganzen Takt
		{
		TCNT0 = T0_PRELOAD;					// in Taktmitte die Bezugszeit zurcksetzen

		bitcount++;
		if (bitcount == 16)  				// Zhler fr eingegangene Bits incrementieren und testen
			{
			ir_start = 0;						// der nchste Impuls startet das Einlesen neu
			ir_ok = 1;							// Flag fr Erfolg
			}

		ir_in = ir_in << 1;					// die Daten um 1 Bit nach links schieben
		if (!(PCINT_PORT & (1<<PCINT_PIN)))// level von PinB2 testen, wenn 1, dann lag eine rising edge vor
			{										// 				 			     wenn 0, dann lag eine falling edge vor
			ir_in++;								// bei falling edge hinten eine 1 anhngen
			}
													// Auslser steigende Flanke -> eine Null anhngen
													// das geht aber beim shiften automatisch
		}
	}
	
if (ir_start == 0)							// wenn Fehler oder Ende erreicht
	{
	TCCR0B = T0_STOP;							// Timer0 stoppen
	only_falling_edge = TRUE;				// auf eine rising edge wird nicht reagiert
													// damit wird die noch folgende rising edge bei einer "1" ignoriert
	}
}

// --- [ end of file ] --------------------------------------------------------
