// --- [ usi_ir_master.c ] ----------------------------------------------------
//       tab = 3, 24.01.09
// Ein ATTINY25 als TWI-Slave steuert ber IR einen IR-Slave
// Die TWI-Adresse ist auf 0x40 (64) eingestellt und kann via General_Call gendert werden
//
// Die IR-Adresse (0-31) ist fest im IR-Slave programmiert.
// TWI-Kommunikation aus der Sicht des TWI-Masters:
//
// Der TWI Slave wird mit drei Byte angesprochen:
// - Byte 1: TWI-Adresse
// - Byte 2: IR-Adresse des IR-Slave (0-31),
// - Byte 3: IR-Befehl   
//				 oberes Nibble -> Gert (Gerte), die geschaltet werden sollen
// -        unteres Nibble -> der dem zu schaltenden Gert zugehrige Sollwert
//
//              Gert Schalterstellung   
//               S3	 S3:   0 = Aus, 1 = Ein
//               |S2  |S2:  0 = Aus, 1 = Ein
//               ||S1 ||S1: 0 = Aus, 1 = Ein
//               |||  |||
// Beispiel:  0b0001_0001  -> Gert.1 wird eingeschaltet
//            0b0011_0010  -> Gert.1 wird ausgeschaltet, Gert.2 wird eingeschaltet !!!
//            0b0111_0101  -> Gert.1 und .3 werden eingeschaltet, Gert.2 ausgeschaltet
// wenn im upper-nibble ein Gert mit 0 markiert ist, dann ist der zugehrige Wert im lower-nibble ohne Bedeutung

// Abfrage des TWI-Slave durch den TWI-Master:
// - Byte 1: TWI_Adresse + 1
// - Byte 2: Anzahl der zu sendenden Bytes (nur 1 Byte macht Sinn!)
//
// Liefert der TWI-Slave den Wert 128 zurck, dann ist eine Transaktion im Gange (Dauer bis 65ms).
//
// Folgende Werte sind Fehlercodes: 255 - Slave hat nicht geantwortet
//                                  254 - der Slave hat eine falsche Adresse zurckgeliefert
// Funktionsabslauf:
//
// Wenn der IR-Master einen Befehls erhlt, dann funkt er den IR-Slave mit dem toggle-Code '000' (also keine
// Vernderung) an (16bit).
// Der IR-Slave sendet darauf hin den aktuellen Bitstatus seiner drei Ports zurck (8bit).
// Darauf stellt der IR-Master fest, welche Bits gendert werden mssen.
// Entsprechend werden toggle-Bits gesetzt und an den IR-Slave zusammen mit seiner Adresse bertragen (16bit).
// Der IR-Slave gibt als Besttigung den neuen Portstatus zurck.
// Stimmt das empfangene Byte mit dem gesendeten im Bereich der Adresse berein, wird dies als Erfolg gewertet.
 
// Diese Besttigung (bzw. die Fehlermeldung) des IR_slaves wird in den TWI-Ausgangspuffer des IR_Master geschrieben und
// kann vom TWI-Master abgeholt werden.

// Whrend der Kommunikation des TWI-Slaves mit dem IR_Slave ist das Bit.7 im Rckgabewert des
// TWI-Slave gesetzt (der TWI-Slave fhrt gerade einen Befehl aus - Dauer bis ca. 65ms).

// Mit dem IR-Befehl 0x00 (bzw. beim Fehlen eines Befehls) wird nur der aktuelle Status des IR-Slaves geholt 
// und in den TWI-Buffer geschrieben. 
// Beispiel: 0x40, 0x10  holt ber den IR-Master mit TWI-Adresse 64 vom IR-Slave 16 den Portstatus
//           0x40, 0x10, 0x73 setzt die Schalter 1 und 2 am IR-Slave 16
//           0x41  befragt den TWI-Slave (IR-MAster) und liefert den aktuellen Portstatus zurck

// IR-Kommunikation:
// Ansteuerung eines IR-Senders, Trger 36KHz, Impulsbreite 500 + 500us = 18 cycles / burst
// benutzt den Timer0 fr die Erzeugung des Trgers auf Pin OC0A,
// Zeitdauer wird ber delay.h geregelt
// Der IR-Master sendet seine 8Bit immer doppelt: 1x in Originalform 1x negiert.
// Der IR-Slave prft, ob beide Byte identisch sind.
// Der IR-Slave selbst sendet nur 1 Byte. Der Master prft, ob die IR-Adresse des Slave korrekt ist.
//
// Der IR-Befehl an den IR-Slave wird vom Programm generiert und
// besteht aus einer 5-Bit Adresse            bbbbb...
// toggle-Bits fr die 3 Schalter xyz         bbbbbxyz
// bzw. 3x 0 fr die Abfrage der Schalter     bbbbb000

// ----------------------------------------------------------------------------

#include <avr/io.h>
#include <avr/interrupt.h>
#include <inttypes.h>
#include <util/delay.h>
#include <stdlib.h>
#include <avr/eeprom.h>
#include "usi_twi_slave.h"
#include "ir_master.h"

#define MAXINPUT 			5					// maximale Anzahl von Eingabebytes bei TWI-read
													// max Anzahl der zu sendenden Bytes siehe usi_twi_slave
													// eigene TWI-Addresse siehe usi_twi_slave.h
// Globale Variablen
uint8_t eemem[3] EEMEM;						// eemem[0] = TWI_ADR, 
													// eemem[1] = OSCCAL
													// eemem[2] = Kalibriermodus (wenn == 1)
volatile uint8_t General_Call;

uint8_t send_to_ir_slave(uint8_t adr, uint8_t cmd);

// ----------------------------------------------------------------------------
// gibt an OC0A = PB.0 fr 15 Sekunden ein 36kHz - Signal aus
// durch Messung der Frequenz und Korrektur des OSCCAL-Wertes (im eeprom) 
// kann auf die Sollfrequenz abgestimmt werden

void kalibrieren(void)
{
uint8_t i,j;

DDR_IR_PORT |= (1<<IR_PIN);				// Sendepin als Ausgang schalten
OCR0A  = COMPARE;								// Comparewert fr toggle OC0A
TCCR0A = IR_CONFIG;							// Register fr die Frequenzerzeugung setzen
TCCR0B = PWM_ON;								// PWM starten, 

j = 150;											// 150 * 1Sek
while(j)
	{
	i = 40;										// 40 * 25ms = 1000ms
	while(i)
		{
		_delay_ms(25);							// max. errreichbare Zeit = 262ms / F_CPU
		i--;
		}
	j--;
	}
	
TCCR0B = PWM_OFF;								// PWM abschalten
DDR_IR_PORT &= ~(1<<IR_PIN);				// Sendepin als Eingang schalten
IR_PORT &= ~(1<<IR_PIN);					// Sendepin auf 0
}


// ----------------------------------------------------------------------------
int main( void )
{
uint8_t j;										// Zhler
uint8_t twi_buffer[2];						// Array um eingegangene TWI-Daten abzuholen

													// Oszillator luft mit 8MHz, Vorteiler 1 fr 8MHz
CLKPR = (1<<CLKPCE);							// 0b10000000 schreiben
CLKPR = (0<<CLKPS0);							// danach sofort den Teiler setzen

// PORTB		PB.0 = SDA
//				PB.1 = OC0B -> IR-Sender
//				PB.2 = SCL
//				PB.3 = frei
//				PB.4 = PCINT -> Empfnger

PORTB|= (1<<PB3);								// die Led signalisiert einen Fehler in der letzten bertragung
DDRB |= (1<<PB3);
													// zum Kalibrieren des Systemtaktes kann ein OSCCAL-Wert
													// ins Eeprom an Pos. 2 geschrieben werden
j = eeprom_read_byte(&eemem[1]);			// wenn er grer als 50 und kleiner als 150 ist,
													// dann wird er bernommen.
if (j < 150 && j > 50) OSCCAL = j; else OSCCAL = OSCCAL_DEFAULT;
													// Achtung, im eeprom an Pos.0 steht die TWI-Adresse !!!!!
													// sie wird automatisch beim Initialisieren des TWI-Modus gelesen
j = eeprom_read_byte(&eemem[2]);			// wenn hier eine 1 steht, dann in den Kalibriermodus wechseln
if (j == 1) kalibrieren();
	
sei();	 										// enable_interrupt

usi_twi_slave_init();						// die TWI_Adresse in in usi_twi_slave.h definiert

#if TESTMODE == FALSE

// ---- das eigentliche Programm ----------------------------------------------

while(1)											// Endlosschleife
	{
	j = 0;
	if (!(usi_is_activ()))					// testen, ob noch einen bertragung am Laufen ist
		{								
		j = usi_twi_rx_bytes();				// wenn nicht, dann prfen, ob Daten eingegangen sind
		}
	
	if(j)											// wenn mindestens 1 Byte bertragen worden ist, dann
		{		
		if (General_Call) 					// wenn ein General_Call vorliegt
			{
			eeprom_write_twi_adr();			// die neue Adresse prfen und ggf. speichern
			General_Call = 0;
			continue;							// den Rest nicht mehr abarbeiten
			}

		if (j > 2) j = 2;						// j begrenzen, es werden nur 2 Byte bentigt

		twi_buffer[0] = 128;					// 128 in den TWI-Ausgangspuffer schreiben,
		usi_twi_put_data(twi_buffer, 1);	// fr den twi-master als Zeichen fr eine aktive bertragung
 
		usi_twi_get_data(twi_buffer, j);	// die eingegangenen Daten abholen und in twi_buffer ablegen
		if (j == 1) twi_buffer[1] = 0;	// wenn der zweite Parameter fehlt, dann eine 0 einsetzen
													// und weiterreichen an die Senderoutine
													// diese umstndliche Byte-Schieberei entstand deswegen, damit
													// die Funktion send_to_ir_slave() auch in anderem Kontext genutzt
													// werden kann.
		twi_buffer[0] = send_to_ir_slave(twi_buffer[0], twi_buffer[1]);

		usi_twi_put_data(twi_buffer, 1);	// slave_status in den TWI_Ausgangs_Buffer schreiben
													// bei Fehlercode (> 7) die Led einschalten
		if (twi_buffer[0] > 0x07) PORTB &= ~(1<<PB3); else PORTB |= (1<<PB3); 
		}
													// sleep einleiten, wenn keine bertragung aktiv
													// und keine Daten vorhanden sind
	if ((!(usi_is_activ()))	&&	(!(usi_twi_rx_bytes())))			
		{
		j = MCUCR;								// Register sichern
		MCUCR = (1<<SE | 1<<SM1);			// powerdown
		asm volatile("sleep");
		MCUCR = j;								// Register zurckschreiben
		}
	}												// while(1)
}

// ---- eine selbstablaufende Sendefunktion, die jede Sekunde einen neuen Wert an den Slave sendet

#else												// Testmodus

#define IR_SLAVE	16							// Adresse
uint8_t i, z;
j = 0x70 + 1;									// das sind die Datenbits (Bit.2..0) + Maske 0x70
													// die Maske legt fest, welche Bits am Slave geschaltet werden soll
while(1)
	{
	z = send_to_ir_slave(IR_SLAVE, j);		// IR_Slave_Adr + cmd senden
	if (z > 0x07) PORTB &= ~(1<<PB3); else  PORTB |= (1<<PB3);
	i = 40;										// 40 * 25ms = 1000ms
	while(i)
		{
		_delay_ms(25);							// max. errreichbare Zeit = 262ms / F_CPU
		i--;
		}

	j++;											// Bitwert (= Schalter) incrementieren
	j &= 0x07;									// auf 7 begrenzen
	j += 0x70;									// die MAske wieder hinzufgen

	}
}

#endif

// ----------------------------------------------------------------------------
// erhlt die IR-Adresse und das Kommando bergeben
// kann auch unabhngig vom TWI-Modul zur Kommunikation mit dem IR_Slave genutzt werden
uint8_t send_to_ir_slave(uint8_t adr, uint8_t cmd)
{													
uint8_t slave_status, x;

if (adr > 31) return(253);					// die Slave-Adresse darf nicht grer als 31 sein

adr *= 8;										// bergebene IR-sLave-Adresse (bis 31) nach links schieben

													// Statusabfrage des IR-Slaves wird in jedem Fall bentigt
													// bei Fehler wird Wert > 7 zurckgegeben
// _delay_ms(3);								// vor dem 1. Senden ist keine Wartezeit erforderlich
slave_status = IR_send(adr);				// die Antwort oder einen Fehlercode abholen
		
													// der Status von Bit.0 ... Bit.2 wird zurckgeliefert
													// der Slave hat nur 3 Schaltkanle
if (slave_status < 8)						// dann wurde kein Fehler gemeldet
	{												// nachdem der Empfnger sich gemeldet hat, kann nun der
													// gewnschte Schalter bettigt werden
													// der Befehl setzt sich aus der Adresse und den Schalterbits zusammen
													// buffer[0] enthlt die aktuelle Schalterstellung des Slaves
											
													// Nachfolgend werden die vorhandenen Schalterstellungen mit den
													// gewnschten verglichen. Sind sie ungleich, dann wird ein
													// toggle-Bit gesetzt, das anschlieend an den Slave gesendet wird
													// es werden also nur die notwendigen Vernderungen gesendet
													
	x = 0;										// x sammelt die toggle-Bits								
	if (cmd & 0x10)							// testen, ob Kanal.1 zu schalten ist, wenn ja, toggle-Bit setzen
		{
		if ((cmd & 0x01) != (slave_status & 0x01)) x = 1;
		}

	if (cmd & 0x20)							// testen, ob Kanal.2 zu schalten ist, wenn ja, toggle-Bit setzen
		{
		if ((cmd & 0x02) != (slave_status & 0x02)) x += 2;
		}

	if (cmd & 0x40)							// testen, ob Kanal.3 zu schalten ist, wenn ja, toggle-Bit setzen
		{
		if ((cmd & 0x04) != (slave_status & 0x04)) x += 4;
		}
													// nur dann zurckschreiben an den Slave,
	if ( x )										// wenn sich wirklich Vernderungen ergeben haben
		{											// twi_buffer[0] enthlt die (Adresse << 3)
		_delay_ms(3);							// kurze Wartezeit bis zum Aktivieren des Senders
		slave_status = IR_send(adr + x); // die Antwort oder den Fehlercode abholen
		}	
	}												// if slave_status < 8)

return(slave_status);						// slave_status enthlt entweder den Pinstatus oder den Fehlercode
}		




// --- [ end of file ] --------------------------------------------------------


